from typing import TYPE_CHECKING, Protocol, Any, Literal, Dict

from yarl import URL

from .tools import request

if TYPE_CHECKING:
    class _ApiCall(Protocol):
        async def __call__(self, **data: Any) -> Any:
            ...


class API:
    def __init__(self, api: str, token: str):
        self.api = URL(api)
        self.headers = {"Authorization": f"Bearer {token}"}

    async def call_api(
            self,
            path: str, method: Literal["GET", "PUT", "POST", "DELETE", "OPTIONS", "HEAD", "PATCH", "TRACE"], **data: Any
    ) -> Any:
        return await request(method, str(self.api / path), self.headers, **data)

    async def system_bulletin(self) -> list:
        """
        获取平台公告
        tag: 系统信息
        code: 200,500
        """
        return await self.call_api("system/bulletin", "GET")

    async def system_clients(self) -> dict:
        """
        获取客户端软件信息
        tag: 系统信息
        code: 200,500
        """
        return await self.call_api("system/clients", "GET")

    async def system_policy(self, type_: Literal["tos", "content", "privacy", "rule", "refund"]) -> dict:
        """
        获取服务条款和策略
        tag: 系统信息
        code: 200,500
        :param type_: <str>(必要) 策略类型
        """
        return await self.call_api("system/policy", "GET", type=type_)

    async def user_info(self) -> dict:
        """
        获取用户基本信息
        tag: 用户信息
        code: 200,500
        """
        return await self.call_api("user/info", "GET")

    async def user_resetToken(self) -> str:
        """
        重置用户访问密钥
        tag: 用户信息
        code: 200,500
        """
        return await self.call_api("user/reset_token", "POST")

    async def user_dataPlans(self, status: Literal["valid", "invalid", "all"] = "valid") -> list:
        """
        获取用户流量包
        tag: 用户信息
        code: 200,500
        :param status: <str>(非必要) 流量包状态
        """
        return await self.call_api("user/data_plans", "GET", status=status)

    async def user_trafficHistory(self, type_: Literal["day", "week", "month"] = "day") -> list:
        """
        获取用户流量历史
        tag: 用户信息
        code: 200,500
        :param type_: <str>(必要) 查询时间类型
        """
        return await self.call_api("user/traffic_history", "GET", type=type_)

    async def nodes(self) -> dict:
        """
        列出所有节点
        tag: 节点管理
        code: 200,500
        """
        return await self.call_api("nodes", "GET")

    async def get_tunnels(self) -> list:
        """
        列出所有隧道
        tag: 隧道管理
        code: 200,500
        """
        return await self.call_api("tunnels", "GET")

    async def post_tunnels(
            self, name: str, type_: Literal["tcp", "udp", "http", "https", "wol", "etcp", "eudp"], note: str,
            extra: str, local_ip: str, local_port: int, remote: str
    ) -> dict:
        """
        创建隧道
        tag: 隧道管理
        code: 201,500
        :param name: <str>(必要) 隧道名称
        :param type_: <str>(必要) 隧道类型
        :param note: <str>(必要) 隧道备注
        :param extra: <str>(必要) 额外参数
        :param local_ip: <str>(必要) 本地 IP
        :param local_port: <int>(必要) 本地端口
        :param remote: <str>(必要) 远程地址(节点的host)
        """
        return await self.call_api(
            "tunnels", "POST", name=name, type=type_, note=note, extra=extra, local_ip=local_ip,
            local_port=local_port, remote=remote
        )

    async def tunnel_config(self, query: str) -> str:
        """
        获取隧道配置文件
        tag: 隧道管理
        code: 200,500
        :param query: <str>(必要) 隧道 ID / 多个隧道ID 用 , 分隔 / n<节点ID>
        """
        return await self.call_api("tunnel/config", "POST", query=query)

    async def tunnel_edit(self, id_: int, note: str, local_ip: str, local_port: int, extra: str) -> dict:
        """
        编辑隧道
        tag: 隧道管理
        code: 200,500
        :param id_: <int>(必要) 隧道 ID
        :param note: <str>(必要) 隧道备注
        :param local_ip: <str>(必要) 本地 IP
        :param local_port: <int>(必要) 本地端口
        :param extra: <str>(必要) 额外参数
        """
        return await self.call_api(
            "tunnel/edit", "POST", id=id_, note=note, local_ip=local_ip, local_port=local_port, extra=extra
        )

    async def tunnel_migrate(self, id_: int, node: int) -> dict:
        """
        迁移隧道
        tag: 隧道管理
        code: 204,500
        :param id_: <int>(必要) 隧道 ID
        :param node: <int>(必要) 节点 ID
        """
        return await self.call_api("tunnel/migrate", "POST", id=id_, node=node)

    async def tunnel_delete(self, ids: str) -> dict:
        """
        删除隧道
        tag: 隧道管理
        code: 200,500
        :param ids: <str>(必要) 隧道 ID / 多个隧道ID 用 , 分隔
        """
        return await self.call_api("tunnel/delete", "POST", ids=ids)

    async def tunnel_traffic(self, id_: int) -> Dict[str, int]:
        """
        获取流量使用记录
        tag: 隧道管理
        code: 200,500
        :param id_: <int>(必要) 隧道 ID
        """
        return await self.call_api("tunnel/traffic", "GET", id=id_)

    async def tunnel_auth(self, id_: int, ip: str) -> str:
        """
        授权 IP 访问隧道
        tag: 访问认证
        code: 200,500
        :param id_: <int>(必要) 隧道 ID
        :param ip: <str>(必要) IP 地址
        """
        return await self.call_api("tunnel/auth", "POST", id=id_, ip=ip)

    async def tunnel_authQuery(self, id_: int, port: int) -> dict:
        """
        根据节点 IP 和端口查询隧道，用于访问认证授权
        tag: 访问认证
        code: 200,500
        :param id_: <int>(必要) 节点 ID
        :param port: <int>(必要) 端口
        """
        return await self.call_api("tunnel/auth_query", "GET", id=id_, port=port)

    async def get_computers(self) -> list:
        """
        获取计算机列表
        tag: 计算机管理
        code: 200,500
        """
        return await self.call_api("computers", "GET")

    async def post_computers(self, name: str, type_: Literal[1, 2], data: str, media: str) -> dict:
        """
        添加计算机
        tag: 计算机管理
        code: 201,500
        :param name: <str>(必要) 计算机名称
        :param type_: <int>(必要) 1:WOL 2:WOL (密码保护)
        :param data: <str>(必要) MAC地址
        :param media: <str>(必要) 节点ID
        """
        return await self.call_api("computers", "POST", name=name, type=type_, data=data, media=media)

    async def patch_computers(self, id_: int, name: str, type_: Literal[1, 2], data: str, media: str) -> str:
        """
        更新计算机信息
        tag: 计算机管理
        code: 204,500
        :param id_: <int>(必要) 计算机ID
        :param name: <str>(必要) 计算机名称
        :param type_: <int>(必要) 1:WOL 2:WOL (密码保护)
        :param data: <str>(必要) MAC地址
        :param media: <str>(必要) 节点ID
        """
        return await self.call_api(
            "computers", "PATCH", id=id_, name=name, type=type_, data=data, media=media
        )

    async def computer_delete(self, id_: int) -> str:
        """
        删除计算机
        tag: 计算机管理
        code: 204,500
        :param id_: <int>(必要) 计算机ID
        """
        return await self.call_api("computer/delete", "POST", id=id_)

    async def computer_poweron(self, id_: int, password: str) -> str:
        """
        执行远程开机
        tag: 计算机管理
        code: 204,500
        :param id_: <int>(必要) 计算机ID
        :param password: <str>(必要) 计算机密码
        """
        return await self.call_api("computer/poweron", "POST", id=id_, password=password)


__all__ = ["API"]



from nonebot.plugin import PluginMetadata, inherit_supported_adapters, require

require("nonebot_plugin_alconna")

from typing_extensions import Annotated
from typing import Literal

from arclet.alconna import Args
from nonebot_plugin_alconna import Alconna, Match, AlconnaMatch

from .api import API
from .tools import (
    send, bytes_to_human, timestamp_to_date, DAY_WEEK_MONTH, TRAFFIC_PLAN_TYPE, wrap_cmd, unescape_html, RawCommand,
    HELPS
)
from .config import config, Config

api = API(config.natfrp_api, config.natfrp_token)


@wrap_cmd(Alconna(config.natfrp_cmd_help, Args["index?", int]))
async def help_cmd_handle(index: Annotated[Match[int], AlconnaMatch("index")]):
    if index.available:
        await send(f"{index.result}.{HELPS[index.result]}")
    else:
        await send("\n".join(f"{i}.{h}" for i, h in enumerate(HELPS)))


@wrap_cmd(Alconna(config.natfrp_cmd_powerOn, Args["id", int]["password?", str, ""]))
async def power_on_cmd_handle(
        id_: Annotated[Match[int], AlconnaMatch("id")], password: Annotated[Match[str], AlconnaMatch("password")]
):
    await send(await api.computer_poweron(id_.result, password.result))


@wrap_cmd(Alconna(config.natfrp_cmd_tunnels))
async def tunnels_cmd_handle(cmd: Annotated[str, RawCommand()]):
    tunnels = await api.get_tunnels()
    msg = cmd + "：\n状态 ID 名称\n"
    for tunnel in tunnels:
        msg += f"{'在线🟢' if tunnel['online'] else '离线🔴'} {tunnel['id']} {tunnel['name']}\n"
    await send(msg)


@wrap_cmd(Alconna(
    config.natfrp_cmd_trafficHistory,
    Args["id?", int]["type?", Literal["day", "week", "month", "日", "周", "月"], "day"]
))
async def traffic_history_cmd_handle(
        cmd: Annotated[str, RawCommand()], id_: Annotated[Match[int], AlconnaMatch("id")],
        type_: Annotated[Match[str], AlconnaMatch("type")]
):
    msg = cmd + "：\n"
    if id_.available:
        history = await api.tunnel_traffic(id_.result)
        msg += "\n".join(f"{timestamp_to_date(x[0])} :\n    {bytes_to_human(x[1])}" for x in history.items())
    else:
        history = await api.user_trafficHistory(type_=DAY_WEEK_MONTH[type_.result])
        msg += "\n".join(f"{x[0]} :\n    {bytes_to_human(x[1])} / {bytes_to_human(x[2])}" for x in history)
    await send(msg)


@wrap_cmd(Alconna(config.natfrp_cmd_trafficPlans, Args[
    "type", Literal["valid", "invalid", "all", "可用", "不可用", "全部"], "all"
]))
async def traffic_plans_cmd_handle(
        cmd: Annotated[str, RawCommand()], type_: Annotated[Match[str], AlconnaMatch("type")]
):
    plans = await api.user_dataPlans(status=TRAFFIC_PLAN_TYPE[type_.result])
    msg = f"{cmd}({type_.result})：\n"
    msg += "\n".join(
        f"{timestamp_to_date(x['start_time'])}~{timestamp_to_date(x['end_time'])}:\n    "
        f"{bytes_to_human(x['remaining'])}/{bytes_to_human(x['total'])}({round(x['remaining'] / x['total'] * 100, 2)}%)"
        for x in plans
    )
    await send(msg)


@wrap_cmd(Alconna(config.natfrp_cmd_userInfo))
async def user_info_cmd_handle():
    u = await api.user_info()
    await send(
        f"-----账号-----\n"
        f"{u['name']}<{u['id']}>\n"
        f"{u['group']['name']}<{u['speed']}>\n"
        f"☆{timestamp_to_date(u['group']['expires'])}☆\n"
        f"-----流量-----\n"
        f"日消耗:{bytes_to_human(u['traffic'][0])}\n"
        f"总剩余:{bytes_to_human(u['traffic'][1])}\n"
        f"-----签到-----\n"
        f"总天数:{u['sign']['days']}\n"
        f"今日签到:{'是' if u['sign']['signed'] else '否'}\n"
        f"上次签到:{u['sign']['last']}\n"
        f"累计获取:{u['sign']['traffic']}GiB"
    )


@wrap_cmd(Alconna(config.natfrp_cmd_announcement, Args["index", int, 0]))
async def announcement_cmd_handle(index: Annotated[Match[int], AlconnaMatch("index")]):
    announcements = await api.system_bulletin()
    index = min(max(len(announcements) - 1, 0), index.result)
    a = announcements[index]
    await send(unescape_html(f"{a['icon']}{a['title']}{a['icon']}\n{a['time']}\n{a['content']}"))


@wrap_cmd(Alconna(config.natfrp_cmd_auth, Args["id", int]["ip", str]))
async def auth_cmd_handle(
        id_: Annotated[Match[int], AlconnaMatch("id")], ip: Annotated[Match[str], AlconnaMatch("ip")]
):
    await send("IP:" + await api.tunnel_auth(id_.result, ip.result))


@wrap_cmd(Alconna(config.natfrp_cmd_showPCs))
async def show_pcs_cmd_handle():
    pcs = await api.get_computers()
    await send("\n".join(f"{pc['id']}.{pc['name']}" + ("<加密>" if pc["type"] != 1 else "") for pc in pcs))


__plugin_meta__ = PluginMetadata(
    name="SakuraFrp",
    description="SakuraFrp管理插件，natfrp.com。",
    usage="\n".join(f"{i}.{h}" for i, h in enumerate(HELPS)),
    type="application",
    homepage="https://github.com/eya46/nonebot_plugin_enatfrp",
    config=Config,
    supported_adapters=inherit_supported_adapters("nonebot_plugin_alconna"),
)



from typing import Optional, List

from nonebot import get_driver
from pydantic import BaseModel, Field, AnyUrl


class Config(BaseModel):
    natfrp_api: AnyUrl = Field(default="https://api.natfrp.com/v4")
    natfrp_token: Optional[str] = Field(default=None)
    natfrp_at: bool = Field(default=False)
    natfrp_reply: bool = Field(default=False)
    natfrp_use_start: bool = Field(default=True)
    natfrp_use_sep: bool = Field(default=False)

    natfrp_cmd_help: List[str] = Field(default=["frp帮助"])
    natfrp_cmd_powerOn: List[str] = Field(default=["开机"])
    natfrp_cmd_tunnels: List[str] = Field(default=["隧道状态"])
    natfrp_cmd_trafficHistory: List[str] = Field(default=["流量", "流量历史"])
    natfrp_cmd_trafficPlans: List[str] = Field(default=["流量包", "流量套餐"])
    natfrp_cmd_userInfo: List[str] = Field(default=["我的信息", "用户信息"])
    natfrp_cmd_announcement: List[str] = Field(default=["公告"])
    natfrp_cmd_auth: List[str] = Field(default=["授权"])
    natfrp_cmd_showPCs: List[str] = Field(default=["计算机列表"])


config: Config = Config(**get_driver().config.dict())

__all__ = ["Config", "config"]



import time
from functools import wraps
from html import unescape
from json import loads
from typing import Union, Any, Callable, Optional, List

from arclet.alconna import Alconna
from nonebot import get_driver
from nonebot.internal.driver import HTTPClientMixin, Request
from nonebot.internal.matcher import current_matcher
from nonebot.internal.params import Depends
from nonebot.permission import SUPERUSER
from nonebot.typing import T_Handler, T_State
from nonebot_plugin_alconna import UniMessage, on_alconna, ALCONNA_RESULT
from nonebot_plugin_alconna.uniseg import Receipt

from .config import config
from .exception import eNatFrpAPIException

if isinstance(driver := get_driver(), HTTPClientMixin):  # type:ignore
    driver: HTTPClientMixin


    async def request(method: str, url: str, headers, **data) -> Any:

        resp = await driver.request(Request(
            method, url,
            headers=headers,
            **({"params": data} if method == "GET" else {"json": data})
        ))

        res = loads(resp.content)
        if resp.status_code == 500:
            raise eNatFrpAPIException(res.get("code", -1), res.get("msg", str(resp.content)))
        return res
else:
    from httpx import AsyncClient


    async def request(method: str, url: str, headers, **data) -> Any:
        async with AsyncClient() as client:
            resp = await client.request(
                method, url, headers=headers,
                **({"params": data} if method == "GET" else {"json": data})
            )
            res = loads(resp.content)
            if resp.status_code == 500:
                raise eNatFrpAPIException(res.get("code", -1), res.get("msg", str(resp.content)))
            return res

# 日_周_月
DAY_WEEK_MONTH = {
    "日": "day",
    "周": "week",
    "月": "month",
    "day": "day",
    "week": "week",
    "month": "month"
}

TRAFFIC_PLAN_TYPE = {
    "可用": "valid",
    "不可用": "invalid",
    "全部": "all",
    "valid": "valid",
    "invalid": "invalid",
    "all": "all"
}

HELPS: List[str] = []


async def send(msg: Union[Any, UniMessage]) -> Receipt:
    if not isinstance(msg, UniMessage):
        msg = UniMessage(msg)
    return await msg.send(at_sender=config.natfrp_at, reply_to=config.natfrp_reply)


def bytes_to_human(data: int) -> str:
    if data < 1024:
        return f"{data}B"
    elif data < 1024 * 1024:
        return f"{round(data / 1024, 1)}KiB"
    elif data < 1024 * 1024 * 1024:
        return f"{round(data / 1024 / 1024, 1)}MiB"
    elif data < 1024 * 1024 * 1024 * 1024:
        return f"{round(data / 1024 / 1024 / 1024, 1)}GiB"
    else:
        return f"{round(data / 1024 / 1024 / 1024 / 1024, 1)}TiB"


# 时间戳转日期
def timestamp_to_date(timestamp: Any) -> str:
    if not isinstance(timestamp, int):
        timestamp = int(timestamp)
    return time.strftime("%Y/%m/%d %Hh", time.localtime(timestamp))


def wrap_cmd(acl: Alconna) -> Callable:
    @wraps(acl)
    def wrapper(func: T_Handler) -> T_Handler:
        @wraps(func)
        async def _catch(*args, **kwargs):
            try:
                await func(*args, **kwargs)
            except eNatFrpAPIException as e:
                await send(
                    f"命令 {_raw_command()} 失败<{e.code}> ->\n"
                    f"{e.message}"
                )

        HELPS.append(acl.get_help().replace("\nUnknown", "").strip())
        on_alconna(
            acl, permission=SUPERUSER, use_cmd_sep=config.natfrp_use_sep, use_cmd_start=config.natfrp_use_start
        ).handle()(_catch)
        return func

    return wrapper


def unescape_html(text: str) -> str:
    return unescape(text)


def _raw_command(state: Optional[T_State] = None) -> str:
    return (state if state else current_matcher.get().state)[ALCONNA_RESULT].result.header_match.result


def RawCommand() -> str:
    return Depends(_raw_command)
